/*
 * ntp_fp.h - definitions for NTP fixed point arithmetic
 */

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include "ntp_types.h"

/*
 * NTP uses two fixed point formats.  The first (l_fp) is the "long" format
 * and is 64 bits with the decimal between 31 and 32 or 
 * 128 bits with the decimal between 63 and 64 for internal computations.  
 * We use the same structure for both  signed  and unsigned values, 
 * which is a big hack but saves rewriting all the operators twice.  
 * Just to confuse this, we also sometimes just 
 * carry the  fractional part in  calculations, in both signed and 
 * unsigned forms. Anyway, an l_fp based on 32 bit longs looks like:
 *
 *    0			  1		      2			  3
 *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |			       Integral Part			     |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |			       Fractional Part			     |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *    0			  1		      2			  3
 * 
 * an l_fp based on 64 bit longs looks like:
 *
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |			       Integral Part			     | 64 bits
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |			       Fractional Part			     | 64 bits
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 * l_fp may be  used for time stamps in the NTP packet header 
 * (in network byte order) on 32 bit machines.  On 64 bit macines
 * a network l_fp is defined as a array of chars.
 *
 * Since network l_fp values most be converted form host to network
 * and network to host byte order macros are defined based on the size of
 * a long handle the two different network representation of l_fp.
 *
 */
typedef struct {
	union {
		U_LONG Xl_ui;
		LONG Xl_i;
	} Ul_i;
	union {
		U_LONG Xl_uf;
		LONG Xl_f;
	} Ul_f;
} l_fp;

#define l_ui	Ul_i.Xl_ui		/* unsigned integral part */
#define	l_i	Ul_i.Xl_i		/* signed integral part */
#define	l_uf	Ul_f.Xl_uf		/* unsigned fractional part */
#define	l_f	Ul_f.Xl_f		/* signed fractional part */

/*
 * The second fixed point format is 32 bits, with the decimal between
 * bits 15 and 16 or 64 bits with the decimal between 31 and 32.  
 * There is a signed version (s_fp) and an unsigned version (u_fp). 
 * The 32 bit fixed point format in network byte order looks like:
 *
 *    0			  1		      2			  3
 *    0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1 2 3 4 5 6 7 8 9 0 1
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *   |		  Integer Part	     |	   Fraction Part	     |
 *   +-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+-+
 *
 * The 32 bit fixed point is used used to represent synchronizing distance
 * and synchronizing dispersion in the NTP packet header (again, in
 * network byte order) and internally to hold both distance and
 * dispersion values (in local byte order).  On 64 bit machines
 * a fixed point number is stored as an array of chars.
 */

typedef LONG s_fp;
typedef U_LONG u_fp;

/*

 The following section contains macros and  parameterize that are used
 to host format numbers to net work numbers.  This section also
 contains the parameterization that are used to compute
 and format l_fp and s_fp numbers.
 
 FP_SECOND              a second in s_fp format.
 FP_SHIFT               shift used to convert from 32 bit fractions
                        64 bit fraction
 FP_MASK                mask to extract fration part of a s_fp
 FP_INT                 mask to extract integer part of a s_fp
 FP_ROUND               round bit position of a s_fp
 SIGN_B                 sign bit of a long
 NET_SHIFT              shift to convert from host fraction to network
                        32 bit fraction
 NTE_FP_SHIFT           shift to convert from netwrok 32 bit fixed point
                        to host fixed point
 SHIFT_UP               converts net fraction to host fraction
 SHIFT_DOWN             converts host fraction to net fraction
 SHIFT_UP_FP            converts net fixed point to host fxied pint (s_fp)
 SHIFT_DOWN_FP          converts host fixed point to net fixed point (s_fp)
 CLOCK_DSCALE_FUDGE     insures that shifts in ../xntpd/ntp_loopfilter.c are
                        positive on 64 bit macines.
*/
#ifdef  LONG8
#define  FP_SECOND           (0x100000000)
#define  FP_SHIFT            32            
#define  FP_MASK             0xFFFFFFFF        
#define  FP_INT              0xFFFFFFFF00000000  
#define  FP_ROUND            0x80000000          	
#define  SIGN_B              0x8000000000000000  	
#define  NET_SHIFT           32		    	
#define  NET_FP_SHIFT        16			
#define  SHIFT_UP(x)         (((U_LONG)(x))<<NET_SHIFT)   
#define  SHIFT_DOWN(x)       (((U_LONG)(x))>>NET_SHIFT)  
#define  SHIFT_UP_FP(x)      (((U_LONG)(x))<<NET_FP_SHIFT)
#define  SHIFT_DOWN_FP(x)    (((U_LONG)(x))>>NET_FP_SHIFT)
#define  CLOCK_DSCALE_FUDGE  13
#define  FRACTION_PREC       (64)
#define  EXTENDI(x) \
        (((x)&0x80000000)?(((LONG)(x))|0xFFFFFFFF00000000):(x))
#define  EXTENDF(x) \
        (((x)&0x800000000000)?(((LONG)(x))|0xFFFF000000000000):(x))
#define  EXTENDC(x) \
          (((x)&0x80)?(((LONG)(x))|0xFFFFFFFFFFFFFF00):(x))
#else
#define	FP_SECOND	     (0x10000)
#define FP_SHIFT       	     16
#define FP_MASK        	     0xFFFF  
#define FP_INT         	     0xFFFF0000
#define FP_ROUND       	     0x8000
#define SIGN_B         	     0x80000000
#define EXTENDI(x)           x
#define EXTENDF(x)           x
#define EXTENDC(x)           (((x)&0x80)?((x)|0xFFFFFF00):(x))
#define SHIFT_UP(x)          (x)
#define SHIFT_DOWN(x)        (x)
#define SHIFT_UP_FP(x)       (x)
#define SHIFT_DOWN_FP(x)     (x)
#define SIGN_B               0x80000000         
#define	FRACTION_PREC	     (32)
#define CLOCK_DSCALE_FUDGE   0
#endif

/*
 * Size of a network long.
 */
#define NET_LONG_LEN 4

#define GET4(x) \
        (((U_LONG)(*(x+0)&0xFF)<<(3*8))|\
         ((U_LONG)(*(x+1)&0xFF)<<(2*8))|\
         ((U_LONG)(*(x+2)&0xFF)<<(1*8))|\
          (U_LONG)(*(x+3)&0xFF)        )

#define PUT4(p, num) \
        *((p)+0) = ((num)>>(32-1*8)); \
        *((p)+1) = ((num)>>(32-2*8)); \
        *((p)+2) = ((num)>>(32-3*8)); \
        *((p)+3) = ((num)          )

#define GET2(x) \
        (((U_LONG)(*(x+0)&0xFF)<<(1*8))|\
          (U_LONG)(*(x+1)&0xFF)        )

#define PUT2(p, num) \
        *((p)+0) = (num)>>8; \
        *((p)+1) = (num)

#if  defined(LONG8)||defined(NET_FP)

#define NET_L_FP_A(x,n) char x[8*n]
#define NET_S_FP_A(x,n) char x[4*n]
#define NET_L_FP(x) NET_L_FP_A(x,1)

/*
 * Byte order conversions
 */
#define normalize_drift(d) \
        ((d)<0?(-((-(d))>>CLOCK_DSCALE_FUDGE)): \
               (    (d   )>>CLOCK_DSCALE_FUDGE))

#define NET_COPY_LF(t, f) memcpy((char *)(t[0]), (char *)(f[0]), 8)

#define  NET_L_CLR(x) memset((char *)x[0], 0, 8)

#define SET_NET_S_FP_A(x, i, v) \
            PUT4(&x[0]+i*4, SHIFT_DOWN_FP(v))

#define SET_NET_L_FP_A(x, i, v) \
        do { U_LONG y; \
             y = (&v)->l_ui;             PUT4(&x[0]+i*8+0, y); \
             y = SHIFT_DOWN((&v)->l_uf); PUT4(&x[0]+i*8+4, y); } while (0)

#define NET_TO_HOST_S_FP_A(x, i) SHIFT_UP_FP(GET4(&x[0]+i*4))

#define NET_TO_HOST_L_FP_A(x, i, v) \
        do {  \
             (&x)->l_ui = EXTENDI( GET4(&v[0]+i*8+0)); \
             (&x)->l_uf = SHIFT_UP(GET4(&v[0]+i*8+4)); \
        } while (0)

#define	HTONS_FP(x)	(htonl(SHIFT_DOWN_FP(x)))

#define HTONL_FP(h, n)  \
        do { U_LONG x; \
             x = (h)->l_ui;             PUT4(n[0]+0, x); \
             x = SHIFT_DOWN((h)->l_uf); PUT4(n[0]+4, x); } while (0)

#define NTOHS_FP(x)   (SHIFT_UP_FP(ntohl(x)))

#define NTOHL_FP(n, h) \
        do {  \
             (h)->l_ui = EXTENDI(GET4(n[0])); \
             (h)->l_uf = SHIFT_UP(GET4(n[0]+4)); \
        } while (0)

#define NTOHL_UFP(n, h) \
        do {  \
             (h)->l_ui = GET4(n[0]); \
             (h)->l_uf = SHIFT_UP(GET4(n[0]+4)); \
        } while (0)

/* funny ones.  Converts ts fractions to net order ts */
#define HTONL_UF(uf, nts) \
        do { U_LONG x = 0; \
             PUT4(nts[0], x); \
             x = SHIFT_DOWN(uf); \
             PUT4(nts[0]+4, x);\
        } while (0)

#define HTONL_F(f, nts) \
	do { 	(nts)->l_uf = EXTENDF(SHIFT_UP_FP(htonl(f))); \
          	if ((f) & SIGN_B) \
               		(nts)->l_i = -1; \
              	else \
               		(nts)->l_i = 0; \
	}

#else
#define NET_L_FP_A(x,n) l_fp x[n]
#define NET_S_FP_A(x, n) s_fp x[n]
#define NET_L_FP(x)  l_fp x
/*
 * Byte order conversions
 */
#define SET_NET_S_FP_A(x, i, v) x[i] = HTONS_FP(v)
#define SET_NET_L_FP_A(x, i, v) HTONL_FP(&v, &x[i])

#define NET_TO_HOST_S_FP_A(x, i) NTOHS_FP(x[i])
#define NET_TO_HOST_L_FP_A(x, i, v) NTOHL_FP(&v[i], &x)

#define normalize_drift(d) 	(d)

#define NET_COPY_LF(t, f) *(t) = *(f)

#define	HTONS_FP(x)	(htonl(x))

#define	HTONL_FP(h, n)	do { (n)->l_ui = htonl((h)->l_ui); \
			     (n)->l_uf = htonl((h)->l_uf); } while (0)

#define	NTOHS_FP(x)	(ntohl(x))

#define NTOHL_UFP(n, h) NTOHL_FP(n, h)

#define	NTOHL_FP(n, h)	do { (h)->l_ui = ntohl((n)->l_ui); \
			     (h)->l_uf = ntohl((n)->l_uf); } while (0)

#define	NTOHL_MFP(ni, nf, hi, hf) \
	do { (hi) = ntohl(ni); (hf) = ntohl(nf); } while (0)

#define	HTONL_MFP(hi, hf, ni, nf) \
	do { (ni) = ntohl(hi); (nf) = ntohl(hf); } while (0)


#define	HTONL_UF(uf, nts) \
	do { (nts)->l_ui = 0; (nts)->l_uf = htonl(uf); } while (0)

#define	HTONL_F(f, nts) do { (nts)->l_uf = htonl(f); \
				if ((f) & 0x80000000) \
					(nts)->l_i = -1; \
				else \
					(nts)->l_i = 0; \
			} while (0)

#define  NET_L_CLR(x) L_CLR(x)

#endif

/*
 * Conversions between the two fixed point types
 */
#define	MFPTOFP(x_i, x_f)	(((x_i)<<FP_SHIFT)|(((x_f)>>FP_SHIFT)&FP_MASK))
#define	LFPTOFP(v)		MFPTOFP((v)->l_ui, (v)->l_uf)

#define UFPTOLFP(x, v) \
	((v)->l_ui = (u_fp)(x)>>FP_SHIFT, (v)->l_uf = (x)<<FP_SHIFT)
#define FPTOLFP(x, v) \
	 (UFPTOLFP((x), (v)), (x) < 0 ? (v)->l_ui -= FP_SECOND : 0)

/*
 * Primitive operations on long fixed point values.  If these are
 * reminiscent of assembler op codes it's only because some may
 * be replaced by inline assembler for particular machines someday.
 * These are the (kind of inefficient) run-anywhere versions.
 */
#define	M_NEG(v_i, v_f) 	/* v = -v */ \
	do { \
		if ((v_f) == 0) \
			(v_i) = -(v_i); \
		else { \
			(v_f) = -(v_f); \
			(v_i) = ~(v_i); \
		} \
	} while(0)

#define	M_NEGM(r_i, r_f, a_i, a_f) 	/* r = -a */ \
	do { \
		if ((a_f) == 0) { \
			(r_f) = 0; \
			(r_i) = -(a_i); \
		} else { \
			(r_f) = -(a_f); \
			(r_i) = ~(a_i); \
		} \
	} while(0)

#define M_ADD(r_i, r_f, a_i, a_f) 	/* r += a */ \
	do { \
		register U_LONG lo_tmp; \
		register U_LONG hi_tmp; \
		\
		lo_tmp = ((r_f) & FP_MASK) + ((a_f) & FP_MASK); \
		hi_tmp = (((r_f) >> FP_SHIFT) & FP_MASK) + \
			 (((a_f) >> FP_SHIFT) & FP_MASK); \
		if (lo_tmp & FP_SECOND) \
			hi_tmp++; \
		(r_f) = ((hi_tmp & FP_MASK) << FP_SHIFT) |\
		         (lo_tmp & FP_MASK); \
		\
		(r_i) += (a_i); \
		if (hi_tmp & FP_SECOND) \
			(r_i)++; \
	} while (0)

#define M_ADD3(r_ovr, r_i, r_f, a_ovr, a_i, a_f) /* r += a, three word */ \
	do { \
		register U_LONG lo_tmp; \
		register U_LONG hi_tmp; \
		\
		lo_tmp = ((r_f) & FP_MASK) + ((a_f) & FP_MASK); \
		hi_tmp = (((r_f) >> FP_SHIFT) & FP_MASK) + \
			 (((a_f) >> FP_SHIFT) & FP_MASK); \
		if (lo_tmp & FP_SECOND) \
			hi_tmp++; \
		(r_f) = ((hi_tmp & FP_MASK) << FP_SHIFT) |\
			 (lo_tmp & FP_MASK); \
		\
		lo_tmp = ((r_i) & FP_MASK) + \
			 ((a_i) & FP_MASK); \
		if (hi_tmp & FP_SECOND) \
			lo_tmp++; \
		hi_tmp = (((r_i) >> FP_SHIFT) & FP_MASK) + \
			 (((a_i) >> FP_SHIFT) & FP_MASK); \
		if (lo_tmp & FP_SECOND) \
			hi_tmp++; \
		(r_i) = ((hi_tmp & FP_MASK) << FP_SHIFT) | \
			 (lo_tmp & FP_MASK); \
		\
		(r_ovr) += (a_ovr); \
		if (hi_tmp & FP_SECOND) \
			(r_ovr)++; \
	} while (0)

#define M_SUB(r_i, r_f, a_i, a_f)	/* r -= a */ \
	do { \
		register U_LONG lo_tmp; \
		register U_LONG hi_tmp; \
		\
		if ((a_f) == 0) { \
			(r_i) -= (a_i); \
		} else { \
			lo_tmp = ((r_f) & FP_MASK) + \
				 ((-(a_f)) & FP_MASK); \
			hi_tmp = (((r_f) >> FP_SHIFT) & FP_MASK) \
			    + (((-(a_f)) >> FP_SHIFT) & FP_MASK); \
			if (lo_tmp & FP_SECOND) \
				hi_tmp++; \
			(r_f) = ((hi_tmp & FP_MASK) << FP_SHIFT) | \
				 (lo_tmp & FP_MASK); \
			\
			(r_i) += ~(a_i); \
			if (hi_tmp & FP_SECOND) \
				(r_i)++; \
		} \
	} while (0)

#define	M_RSHIFTU(v_i, v_f)		/* v >>= 1, v is unsigned */ \
	do { \
		(v_f) = (U_LONG)(v_f) >> 1; \
		if ((v_i) & 01) \
			(v_f) |= SIGN_B; \
		(v_i) = (U_LONG)(v_i) >> 1; \
	} while (0)

#define	M_RSHIFT(v_i, v_f)		/* v >>= 1, v is signed */ \
	do { \
		(v_f) = (U_LONG)(v_f) >> 1; \
		if ((v_i) & 01) \
			(v_f) |= SIGN_B; \
		if ((v_i) & SIGN_B) \
			(v_i) = ((v_i) >> 1) | SIGN_B; \
		else \
			(v_i) = (v_i) >> 1; \
	} while (0)

#define	M_LSHIFT(v_i, v_f)		/* v <<= 1 */ \
	do { \
		(v_i) <<= 1; \
		if ((v_f) & SIGN_B) \
			(v_i) |= 0x1; \
		(v_f) <<= 1; \
	} while (0)

#define	M_LSHIFT3(v_ovr, v_i, v_f)	/* v <<= 1, with overflow */ \
	do { \
		(v_ovr) <<= 1; \
		if ((v_i) & SIGN_B) \
			(v_ovr) |= 0x1; \
		(v_i) <<= 1; \
		if ((v_f) & SIGN_B) \
			(v_i) |= 0x1; \
		(v_f) <<= 1; \
	} while (0)

#define	M_ADDUF(r_i, r_f, uf) 		/* r += uf, uf is U_LONG fraction */ \
	M_ADD((r_i), (r_f), 0, (uf))	/* let optimizer worry about it */

#define	M_SUBUF(r_i, r_f, uf)		/* r -= uf, uf is U_LONG fraction */ \
	M_SUB((r_i), (r_f), 0, (uf))	/* let optimizer worry about it */

#define	M_ADDF(r_i, r_f, f)		/* r += f, f is a LONG fraction */ \
	do { \
		if ((f) > 0) \
			M_ADD((r_i), (r_f), 0, (f)); \
		else if ((f) < 0) \
			M_ADD((r_i), (r_f), (-1), (f));\
	} while(0)

#define	M_ISNEG(v_i, v_f) 		/* v < 0 */ \
	(((v_i) & SIGN_B) != 0)

#define	M_ISHIS(a_i, a_f, b_i, b_f)	/* a >= b unsigned */ \
	(((U_LONG)(a_i)) > ((U_LONG)(b_i)) || \
	  ((a_i) == (b_i) && ((U_LONG)(a_f)) >= ((U_LONG)(b_f))))

#define	M_ISGEQ(a_i, a_f, b_i, b_f)	/* a >= b signed */ \
	(((LONG)(a_i)) > ((LONG)(b_i)) || \
	  ((a_i) == (b_i) && ((U_LONG)(a_f)) >= ((U_LONG)(b_f))))

#define	M_ISEQU(a_i, a_f, b_i, b_f)	/* a == b unsigned */ \
	((a_i) == (b_i) && (a_f) == (b_f))

/*
 * Operations on the long fp format
 */
#define	L_ADD(r, a)	M_ADD((r)->l_ui, (r)->l_uf, (a)->l_ui, (a)->l_uf)
#define	L_SUB(r, a)	M_SUB((r)->l_ui, (r)->l_uf, (a)->l_ui, (a)->l_uf)
#define	L_NEG(v)	M_NEG((v)->l_ui, (v)->l_uf)
#define L_ADDUF(r, uf)	M_ADDUF((r)->l_ui, (r)->l_uf, (uf))
#define L_SUBUF(r, uf)	M_SUBUF((r)->l_ui, (r)->l_uf, (uf))
#define	L_ADDF(r, f)	M_ADDF((r)->l_ui, (r)->l_uf, (f))
#define	L_RSHIFT(v)	M_RSHIFT((v)->l_i, (v)->l_uf)
#define	L_RSHIFTU(v)	M_RSHIFT((v)->l_ui, (v)->l_uf)
#define	L_LSHIFT(v)	M_LSHIFT((v)->l_ui, (v)->l_uf)
#define	L_CLR(v)	((v)->l_ui = (v)->l_uf = 0)

#define	L_ISNEG(v)	(((v)->l_ui & SIGN_B) != 0)
#define L_ISZERO(v)	((v)->l_ui == 0 && (v)->l_uf == 0)
#define	L_ISHIS(a, b)	((a)->l_ui > (b)->l_ui || \
			  ((a)->l_ui == (b)->l_ui && (a)->l_uf >= (b)->l_uf))
#define	L_ISGEQ(a, b)	((a)->l_i > (b)->l_i || \
			  ((a)->l_i == (b)->l_i && (a)->l_uf >= (b)->l_uf))
#define	L_ISEQU(a, b)	M_ISEQU((a)->l_ui, (a)->l_uf, (b)->l_ui, (b)->l_uf)

extern	char *	dofptoa		P((u_fp, int, int, int));
extern	char *	dolfptoa	P((u_long, u_long, int, int, int));

extern	int	atolfp		P((const char *, l_fp *));
extern	int	buftvtots	P((const char *, l_fp *));
extern	void	gettstamp	P((l_fp *));
extern	char *	fptoa		P((s_fp, int));
extern	char *	fptoms		P((s_fp, int));
extern	char *	fptoms		P((s_fp, int));
extern	int	hextolfp	P((const char *, l_fp *));
extern	int	mstolfp		P((const char *, l_fp *));
extern	char *	prettydate	P((l_fp *));
extern	char *	uglydate	P((l_fp *));

extern	void	get_systime	P((l_fp *));
extern	int	step_systime	P((l_fp *));
extern  int 	step_systime_real P((l_fp *));
extern	int	adj_systime	P((l_fp *));

#define	lfptoa(_fpv, _ndec)	mfptoa((_fpv)->l_ui, (_fpv)->l_uf, (_ndec))
#define	lfptoms(_fpv, _ndec)	mfptoms((_fpv)->l_ui, (_fpv)->l_uf, (_ndec))

#define	ntoa(_sin)		numtoa((_sin)->sin_addr.s_addr)
#define	ntohost(_sin)		numtohost((_sin)->sin_addr.s_addr)

#define	ufptoa(_fpv, _ndec)	dofptoa((_fpv), 0, (_ndec), 0)
#define	ufptoms(_fpv, _ndec)	dofptoa((_fpv), 0, (_ndec), 1)
#define	ulfptoa(_fpv, _ndec)	dolfptoa((_fpv)->l_ui, (_fpv)->l_uf, 0, (_ndec), 0)
#define	ulfptoms(_fpv, _ndec)	dolfptoa((_fpv)->l_ui, (_fpv)->l_uf, 0, (_ndec), 1)
#define	umfptoa(_fpi, _fpf, _ndec) dolfptoa((_fpi), (_fpf), 0, (_ndec), 0)
